@@ -821,6 +821,27 @@ static const char *hextab = "0123456789abcdef";
  *  double-check the parsed value, to be sure it's legal for that
  *  type (length, etc.)
  */
+static uint32_t getint(const char *value, char **end)
+{
+	if ((value[0] == '0') && (value[1] == 'x')) {
+		return strtoul(value, end, 16);
+	}
+
+	return strtoul(value, end, 10);
+}
+
+static int check_for_whitespace(const char *value)
+{
+	while (*value) {
+		if (!isspace((int) *value)) return 0;
+
+		value++;
+	}
+
+	return 1;
+}
+
+
 VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 {
 	char		*p, *s=0;
@@ -829,6 +850,8 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 	size_t		length;
 	DICT_VALUE	*dval;
 
+	if (!value) return NULL;
+
 	/*
 	 *	Even for integers, dates and ip addresses we
 	 *	keep the original string in vp->vp_strvalue.
@@ -939,73 +962,48 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			break;
 
 		case PW_TYPE_BYTE:
-			if (value && (value[0] == '0') && (value[1] == 'x')) {
-				goto do_octets;
-			}
+			vp->length = 1;
 
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
+			vp->vp_integer = getint(value, &p);
 			if (!*p) {
 				if (vp->vp_integer > 255) {
 					fr_strerror_printf("Byte value \"%s\" is larger than 255", value);
 					return NULL;
 				}
-				vp->length = 1;
 				break;
 			}
-
-			/*
-			 *	Look for the named value for the given
-			 *	attribute.
-			 */
-			if ((dval = dict_valbyname(vp->attribute, value)) == NULL) {
-				fr_strerror_printf("Unknown value %s for attribute %s",
-					   value, vp->name);
-				return NULL;
-			}
-			vp->vp_integer = dval->value;
-			vp->length = 1;
-			break;
+			if (check_for_whitespace(p)) break;
+			goto check_for_value;
 
 		case PW_TYPE_SHORT:
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
+			vp->vp_integer = getint(value, &p);
+			vp->length = 2;
 			if (!*p) {
 				if (vp->vp_integer > 65535) {
 					fr_strerror_printf("Byte value \"%s\" is larger than 65535", value);
 					return NULL;
 				}
-				vp->length = 2;
 				break;
 			}
-
-			/*
-			 *	Look for the named value for the given
-			 *	attribute.
-			 */
-			if ((dval = dict_valbyname(vp->attribute, value)) == NULL) {
-				fr_strerror_printf("Unknown value %s for attribute %s",
-					   value, vp->name);
-				return NULL;
-			}
-			vp->vp_integer = dval->value;
-			vp->length = 2;
-			break;
+			if (check_for_whitespace(p)) break;
+			goto check_for_value;
 
 		case PW_TYPE_INTEGER:
 			/*
 			 *	Note that ALL integers are unsigned!
 			 */
-			vp->vp_integer = (uint32_t) strtoul(value, &p, 10);
-			if (!*p) {
-				vp->length = 4;
-				break;
-			}
+			vp->vp_integer = getint(value, &p);
+			vp->length = 4;
+			if (!*p) break;
+			if (check_for_whitespace(p)) break;
 
+	check_for_value:
 			/*
 			 *	Look for the named value for the given
 			 *	attribute.
@@ -1016,7 +1014,6 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 				return NULL;
 			}
 			vp->vp_integer = dval->value;
-			vp->length = 4;
 			break;
 
 		case PW_TYPE_DATE:
@@ -1059,8 +1056,8 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			 *	then fall through to raw octets, so that
 			 *	the user can at least make them by hand...
 			 */
-#endif
 	do_octets:
+#endif
 			/* raw octets: 0x01020304... */
 		case PW_TYPE_OCTETS:
 			if (strncasecmp(value, "0x", 2) == 0) {
@@ -1203,7 +1200,7 @@ VALUE_PAIR *pairparsevalue(VALUE_PAIR *vp, const char *value)
 			break;
 
 		case PW_TYPE_TLV: /* don't use this! */
-			if (strncasecmp(value, "0t", 2) != 0) {
+			if (strncasecmp(value, "0x", 2) != 0) {
 				fr_strerror_printf("Invalid TLV specification");
 				return NULL;
 			}
@@ -1261,7 +1258,7 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		return NULL;
 	}
 
-	attr = vendor = 0;
+	vendor = 0;
 
 	/*
 	 *	Pull off vendor prefix first.
@@ -1372,6 +1369,9 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		return NULL;
 	}
 
+	vp->operator = (operator == 0) ? T_OP_EQ : operator;
+	if (!value) return vp;
+
 	size = strlen(value + 2);
 
 	/*
@@ -1414,8 +1414,6 @@ static VALUE_PAIR *pairmake_any(const char *attribute, const char *value,
 		break;
 	}
        
-	vp->operator = (operator == 0) ? T_OP_EQ : operator;
-
 	return vp;
 }
 
@@ -1490,13 +1488,12 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 	if (value && (*value == ':' && da->flags.has_tag)) {
 	        /* If we already found a tag, this is invalid */
 	        if(found_tag) {
-		        pairbasicfree(vp);
 			fr_strerror_printf("Duplicate tag %s for attribute %s",
 				   value, vp->name);
 			DEBUG("Duplicate tag %s for attribute %s\n",
 				   value, vp->name);
+		        pairbasicfree(vp);
 			return NULL;
-
 		}
 	        /* Colon found and attribute allows a tag */
 	        if (value[1] == '*' && value[2] == ':') {
@@ -1539,6 +1536,15 @@ VALUE_PAIR *pairmake(const char *attribute, const char *value, int operator)
 		 */
 	case T_OP_REG_EQ:	/* =~ */
 	case T_OP_REG_NE:	/* !~ */
+		if (!value) {
+			fr_strerror_printf("No regular expression found in %s",
+					   vp->name);
+		        pairbasicfree(vp);
+			return NULL;
+		}
+	  
+		strlcpy(vp->vp_strvalue, value, sizeof(vp->vp_strvalue));
+		vp->length = strlen(vp->vp_strvalue);
 		/*
 		 *	If anything goes wrong, this is a run-time error,
 		 *	not a compile-time error.
@@ -1635,7 +1641,7 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
 	/*
 	 *	We may have Foo-Bar:= stuff, so back up.
 	 */
-	if (attr[len - 1] == ':') {
+	if ((len > 0) && (attr[len - 1] == ':')) {
 		p--;
 		len--;
 	}
@@ -1701,16 +1707,44 @@ VALUE_PAIR *pairread(const char **ptr, FR_TOKEN *eol)
 			vp->flags.do_xlat = 1;
 			vp->length = 0;
 		} else {
-		case T_SINGLE_QUOTED_STRING:
-			vp = pairmake(attr, NULL, token);
-			if (vp) {
-				strlcpy(vp->vp_strvalue, value,
-					sizeof(vp->vp_strvalue));
-				vp->length = strlen(vp->vp_strvalue);
+			/*
+			 *	Parse && escape it, as defined by the
+			 *	data type.
+			 */
+			vp = pairmake(attr, value, token);
+			if (!vp) {
+				*eol = T_OP_INVALID;
+				return NULL;
 			}
 		}
 		break;
 
+	case T_SINGLE_QUOTED_STRING:
+		vp = pairmake(attr, NULL, token);
+		if (!vp) {
+			*eol = T_OP_INVALID;
+			return NULL;
+		}
+
+		/*
+		 *	String and octet types get copied verbatim.
+		 */
+		if ((vp->type == PW_TYPE_STRING) ||
+		    (vp->type == PW_TYPE_OCTETS)) {
+			strlcpy(vp->vp_strvalue, value,
+				sizeof(vp->vp_strvalue));
+			vp->length = strlen(vp->vp_strvalue);
+
+			/*
+			 *	Everything else gets parsed: it's
+			 *	DATA, not a string!
+			 */
+		} else if (!pairparsevalue(vp, value)) {
+			pairfree(&vp);
+			*eol = T_OP_INVALID;
+			return NULL;
+		}
+		break;
 
 		/*
 		 *	Mark the pair to be allocated later.
@@ -1852,6 +1886,8 @@ VALUE_PAIR *readvp2(FILE *fp, int *pfiledone, const char *errprefix)
  *	e.g. "foo" != "bar"
  *
  *	Returns true (comparison is true), or false (comparison is not true);
+ *
+ *	FIXME: Ignores tags!
  */
 int paircmp(VALUE_PAIR *one, VALUE_PAIR *two)
 {
