@@ -133,8 +133,10 @@ typedef struct THREAD_POOL {
 	int cleanup_delay;
 	int spawn_flag;
 
+#ifdef WNOHANG
 	pthread_mutex_t	wait_mutex;
 	fr_hash_table_t *waiters;
+#endif
 
 	/*
 	 *	All threads wait on this semaphore, for requests
@@ -234,7 +236,7 @@ static int setup_ssl_mutexes(void)
 }
 #endif
 
-
+#ifdef WNOHANG
 /*
  *	We don't want to catch SIGCHLD for a host of reasons.
  *
@@ -261,12 +263,13 @@ static void reap_children(void)
 	pthread_mutex_lock(&thread_pool.wait_mutex);
 
 	do {
+	retry:
 		pid = waitpid(0, &status, WNOHANG);
 		if (pid <= 0) break;
 
 		mytf.pid = pid;
 		tf = fr_hash_table_finddata(thread_pool.waiters, &mytf);
-		if (!tf) continue;
+		if (!tf) goto retry;
 
 		tf->status = status;
 		tf->exited = 1;
@@ -274,6 +277,9 @@ static void reap_children(void)
 
 	pthread_mutex_unlock(&thread_pool.wait_mutex);
 }
+#else
+#define reap_children()
+#endif /* WNOHANG */
 
 /*
  *	Add a request to the list of waiting requests.
@@ -367,6 +373,7 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 		entry->request->child_state = REQUEST_DONE;
 		thread_pool.num_queued--;
 		free(entry);
+		entry = NULL;
 	}
 
 	start = 0;
@@ -394,6 +401,7 @@ static int request_dequeue(REQUEST **request, RAD_REQUEST_FUNP *fun)
 	*request = entry->request;
 	*fun = entry->fun;
 	free(entry);
+	entry = NULL;
 
 	rad_assert(*request != NULL);
 	rad_assert((*request)->magic == REQUEST_MAGIC);
@@ -585,7 +593,6 @@ static THREAD_HANDLE *spawn_thread(time_t now)
 	memset(handle, 0, sizeof(THREAD_HANDLE));
 	handle->prev = NULL;
 	handle->next = NULL;
-	handle->pthread_id = NO_SUCH_CHILD_PID;
 	handle->thread_num = thread_pool.max_thread_num++;
 	handle->request_count = 0;
 	handle->status = THREAD_RUNNING;
@@ -664,6 +671,7 @@ int total_active_threads(void)
 }
 
 
+#ifdef WNOHANG
 static uint32_t pid_hash(const void *data)
 {
 	const thread_fork_t *tf = data;
@@ -678,6 +686,7 @@ static int pid_cmp(const void *one, const void *two)
 
 	return (a->pid - b->pid);
 }
+#endif
 
 /*
  *	Allocate the thread pool, and seed it with an initial number
@@ -685,7 +694,7 @@ static int pid_cmp(const void *one, const void *two)
  *
  *	FIXME: What to do on a SIGHUP???
  */
-int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
+int thread_pool_init(CONF_SECTION *cs, int *spawn_flag)
 {
 	int		i, rcode;
 	CONF_SECTION	*pool_cf;
@@ -693,50 +702,48 @@ int thread_pool_init(CONF_SECTION *cs, int spawn_flag)
 
 	now = time(NULL);
 
-	/*
-	 *	We're not spawning new threads, don't do
-	 *	anything.
-	 */
-	if (!spawn_flag) return 0;
-
-	/*
-	 *	After a SIGHUP, we don't over-write the previous values.
-	 */
-	if (!pool_initialized) {
-		/*
-		 *	Initialize the thread pool to some reasonable values.
-		 */
-		memset(&thread_pool, 0, sizeof(THREAD_POOL));
-		thread_pool.head = NULL;
-		thread_pool.tail = NULL;
-		thread_pool.total_threads = 0;
-		thread_pool.max_thread_num = 1;
-		thread_pool.cleanup_delay = 5;
-		thread_pool.spawn_flag = spawn_flag;
-
-		if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
-			radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
-			       strerror(errno));
-			return -1;
-		}
-
-		/*
-		 *	Create the hash table of child PID's
-		 */
-		thread_pool.waiters = fr_hash_table_create(pid_hash,
-							     pid_cmp,
-							     free);
-		if (!thread_pool.waiters) {
-			radlog(L_ERR, "FATAL: Failed to set up wait hash");
-			return -1;
-		}
-	}
+	rad_assert(spawn_flag != NULL);
+	rad_assert(*spawn_flag == TRUE);
+	rad_assert(pool_initialized == FALSE); /* not called on HUP */
 
 	pool_cf = cf_subsection_find_next(cs, NULL, "thread");
-	if (!pool_cf) {
-		radlog(L_ERR, "FATAL: Attempting to start in multi-threaded mode with no thread configuration in radiusd.conf");
+	if (!pool_cf) *spawn_flag = FALSE;
+
+	/*
+	 *	Initialize the thread pool to some reasonable values.
+	 */
+	memset(&thread_pool, 0, sizeof(THREAD_POOL));
+	thread_pool.head = NULL;
+	thread_pool.tail = NULL;
+	thread_pool.total_threads = 0;
+	thread_pool.max_thread_num = 1;
+	thread_pool.cleanup_delay = 5;
+	thread_pool.spawn_flag = *spawn_flag;
+	
+	/*
+	 *	Don't bother initializing the mutexes or
+	 *	creating the hash tables.  They won't be used.
+	 */
+	if (!*spawn_flag) return 0;
+	
+#ifdef WNOHANG
+	if ((pthread_mutex_init(&thread_pool.wait_mutex,NULL) != 0)) {
+		radlog(L_ERR, "FATAL: Failed to initialize wait mutex: %s",
+		       strerror(errno));
 		return -1;
 	}
+	
+	/*
+	 *	Create the hash table of child PID's
+	 */
+	thread_pool.waiters = fr_hash_table_create(pid_hash,
+						   pid_cmp,
+						   free);
+	if (!thread_pool.waiters) {
+		radlog(L_ERR, "FATAL: Failed to set up wait hash");
+		return -1;
+	}
+#endif
 
 	if (cf_section_parse(pool_cf, NULL, thread_config) < 0) {
 		return -1;
@@ -834,12 +841,14 @@ int thread_pool_addrequest(REQUEST *request, RAD_REQUEST_FUNP fun)
 	if (!thread_pool.spawn_flag) {
 		radius_handle_request(request, fun);
 
+#ifdef WNOHANG
 		/*
 		 *	Requests that care about child process exit
 		 *	codes have already either called
 		 *	rad_waitpid(), or they've given up.
 		 */
 		wait(NULL);
+#endif
 		return 1;
 	}
 
@@ -1020,6 +1029,7 @@ static void thread_pool_manage(time_t now)
 }
 
 
+#ifdef WNOHANG
 /*
  *	Thread wrapper for fork().
  */
@@ -1055,6 +1065,7 @@ pid_t rad_fork(void)
 		if (!rcode) {
 			radlog(L_ERR, "Failed to store PID, creating what will be a zombie process %d",
 			       (int) child_pid);
+			free(tf);
 		}
 	}
 
@@ -1108,6 +1119,11 @@ pid_t rad_waitpid(pid_t pid, int *status)
 
 	return 0;
 }
+#else
+/*
+ *	No rad_fork or rad_waitpid
+ */
+#endif
 
 void thread_pool_lock(void)
 {
