@@ -84,10 +84,20 @@ static int eap_handler_cmp(const void *a, const void *b)
 	if (one->eap_id < two->eap_id) return -1;
 	if (one->eap_id > two->eap_id) return +1;
 
-	rcode = fr_ipaddr_cmp(&one->src_ipaddr, &two->src_ipaddr);
+	rcode = memcmp(one->state, two->state, sizeof(one->state));
 	if (rcode != 0) return rcode;
 
-	return memcmp(one->state, two->state, sizeof(one->state));
+	/*
+	 *	As of 2.1.8, we don't key off of source IP.  This
+	 *	a NAS to send packets load-balanced (or fail-over)
+	 *	across multiple intermediate proxies, and still have
+	 *	EAP work.
+	 */
+	if (fr_ipaddr_cmp(&one->src_ipaddr, &two->src_ipaddr) != 0) {
+		DEBUG("WARNING: EAP packets are arriving from two different upstream servers.  Has there been a proxy fail-over?");
+	}
+
+	return 0;
 }
 
 
@@ -283,6 +293,7 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		return RLM_MODULE_INVALID;
 	}
 
+#ifdef WITH_PROXY
 	/*
 	 *	If we're doing horrible tunneling work, remember it.
 	 */
@@ -301,8 +312,9 @@ static int eap_authenticate(void *instance, REQUEST *request)
 
 		return RLM_MODULE_HANDLED;
 	}
+#endif
 
-
+#ifdef WITH_PROXY
 	/*
 	 *	Maybe the request was marked to be proxied.  If so,
 	 *	proxy it.
@@ -348,6 +360,7 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		RDEBUG2("  Tunneled session will be proxied.  Not doing EAP.");
 		return RLM_MODULE_HANDLED;
 	}
+#endif
 
 	/*
 	 *	We are done, wrap the EAP-request in RADIUS to send
@@ -411,8 +424,11 @@ static int eap_authenticate(void *instance, REQUEST *request)
 		 */
 		vp = pairfind(request->reply->vps, PW_USER_NAME);
 		if (!vp) {
-			vp = pairmake("User-Name", request->username->vp_strvalue,
+			vp = pairmake("User-Name", "",
 				      T_OP_EQ);
+			strlcpy(vp->vp_strvalue, request->username->vp_strvalue,
+				sizeof(vp->vp_strvalue));
+			vp->length = request->username->length;
 			rad_assert(vp != NULL);
 			pairadd(&(request->reply->vps), vp);
 		}
@@ -444,12 +460,14 @@ static int eap_authorize(void *instance, REQUEST *request)
 
 	inst = (rlm_eap_t *)instance;
 
+#ifdef WITH_PROXY
 	/*
 	 *	We don't do authorization again, once we've seen the
 	 *	proxy reply (or the proxied packet)
 	 */
 	if (request->proxy != NULL)
                 return RLM_MODULE_NOOP;
+#endif
 
 	/*
 	 *	For EAP_START, send Access-Challenge with EAP Identity
@@ -489,6 +507,8 @@ static int eap_authorize(void *instance, REQUEST *request)
 	    (vp->vp_integer != PW_AUTHTYPE_REJECT)) {
 		vp = pairmake("Auth-Type", inst->xlat_name, T_OP_EQ);
 		if (!vp) {
+			RDEBUG2("Failed to create Auth-Type %s: %s\n",
+				inst->xlat_name, fr_strerror());
 			return RLM_MODULE_FAIL;
 		}
 		pairadd(&request->config_items, vp);
@@ -499,6 +519,8 @@ static int eap_authorize(void *instance, REQUEST *request)
 	return RLM_MODULE_UPDATED;
 }
 
+
+#ifdef WITH_PROXY
 /*
  *	If we're proxying EAP, then there may be magic we need
  *	to do.
@@ -510,6 +532,11 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 	VALUE_PAIR	*vp;
 	EAP_HANDLER	*handler;
 
+	/*
+	 *	Just in case the admin lists EAP in post-proxy-type Fail.
+	 */
+	if (!request->proxy_reply) return RLM_MODULE_NOOP;
+
 	/*
 	 *	If there was a handler associated with this request,
 	 *	then it's a tunneled request which was proxied...
@@ -548,7 +575,7 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 		 *	We are done, wrap the EAP-request in RADIUS to send
 		 *	with all other required radius attributes
 		 */
-		rcode = eap_compose(handler);
+		eap_compose(handler);
 
 		/*
 		 *	Add to the list only if it is EAP-Request, OR if
@@ -592,7 +619,6 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 		RDEBUG2("No pre-existing handler found");
 	}
 
-
 	/*
 	 *	There may be more than one Cisco-AVPair.
 	 *	Ensure we find the one with the LEAP attribute.
@@ -656,7 +682,7 @@ static int eap_post_proxy(void *inst, REQUEST *request)
 
 	return RLM_MODULE_UPDATED;
 }
-
+#endif
 
 /*
  *	The module name should be the only globally exported symbol.
@@ -675,7 +701,11 @@ module_t rlm_eap = {
 		NULL,			/* accounting */
 		NULL,			/* checksimul */
 		NULL,			/* pre-proxy */
+#ifdef WITH_PROXY
 		eap_post_proxy,		/* post-proxy */
+#else
+		NULL,
+#endif
 		NULL			/* post-auth */
 	},
 };
